import logging
import time

from scrapy.dupefilters import BaseDupeFilter
from scrapy.utils.request import request_fingerprint

from aioscrapy.db import get_pool


logger = logging.getLogger(__name__)


class RFPDupeFilter(BaseDupeFilter):
    """ 使用redis集合构建的过滤器"""

    logger = logger

    def __init__(self, server, key, debug=False):
        self.server = server
        self.key = key
        self.debug = debug
        self.logdupes = True

    @classmethod
    async def from_spider(cls, spider):
        settings = spider.settings
        server = get_pool('redis')
        dupefilter_key = settings.get("SCHEDULER_DUPEFILTER_KEY",  '%(spider)s:bloomfilter')
        key = dupefilter_key % {'spider': spider.name}
        debug = settings.getbool('DUPEFILTER_DEBUG', False)
        instance = cls(server, key=key, debug=debug)
        return instance

    @classmethod
    async def from_settings(cls, settings):
        server = get_pool('redis')
        key = settings['DUPEFILTER_KEY'] % {'timestamp': int(time.time())}
        debug = settings.getbool('DUPEFILTER_DEBUG')
        return cls(server, key=key, debug=debug)

    @classmethod
    async def from_crawler(cls, crawler):
        return await cls.from_settings(crawler.settings)

    async def request_seen(self, request):
        fp = self.request_fingerprint(request)
        return await self.server.sadd(self.key, fp) == 0

    def request_fingerprint(self, request):
        return request_fingerprint(request)

    async def close(self, reason=''):
        await self.clear()

    async def clear(self):
        await self.server.delete(self.key)

    def log(self, request, spider):
        if self.debug:
            msg = "Filtered duplicate request: %(request)s"
            self.logger.debug(msg, {'request': request}, extra={'spider': spider})
        elif self.logdupes:
            msg = ("Filtered duplicate request %(request)s"
                   " - no more duplicates will be shown"
                   " (see DUPEFILTER_DEBUG to show all duplicates)")
            self.logger.debug(msg, {'request': request}, extra={'spider': spider})
            self.logdupes = False


class HashMap(object):
    def __init__(self, m, seed):
        self.m = m
        self.seed = seed

    def hash(self, value):
        """
        Hash Algorithm
        :param value: Value
        :return: Hash Value
        """
        ret = 0
        for i in range(len(value)):
            ret += self.seed * ret + ord(value[i])
        return (self.m - 1) & ret


class BloomFilter(object):
    def __init__(self, server, key, bit=30, hash_number=6):
        """
        Initialize BloomFilter
        :param server: Redis Server
        :param key: BloomFilter Key
        :param bit: m = 2 ^ bit
        :param hash_number: the number of hash function
        """
        # default to 1 << 30 = 10,7374,1824 = 2^30 = 128MB, max filter 2^30/hash_number = 1,7895,6970 fingerprints
        self.m = 1 << bit
        self.seeds = range(hash_number)
        self.server = server
        self.key = key
        self.maps = [HashMap(self.m, seed) for seed in self.seeds]

    async def exists(self, value):
        if not value:
            return False
        exist = True
        for map_ in self.maps:
            offset = map_.hash(value)
            exist = exist & await self.server.getbit(self.key, offset)
            if not exist:
                return False
        return exist

    async def insert(self, value):
        """
        add value to bloom
        :param value:
        :return:
        """
        for f in self.maps:
            offset = f.hash(value)
            await self.server.setbit(self.key, offset, 1)


class BloomDupeFilter(RFPDupeFilter):
    """ 使用redis的位图构建的布隆过滤器 """

    def __init__(self, server, key, debug, bit, hash_number):
        super().__init__(server, key, debug)
        self.bit = bit
        self.hash_number = hash_number
        self.bf = BloomFilter(server, self.key, bit, hash_number)

    @classmethod
    async def from_settings(cls, settings):
        server = await get_pool('redis')

        key = settings.get('DUPEFILTER_KEY', 'dupefilter:%(timestamp)s') % {'timestamp': int(time.time())}
        debug = settings.getbool('DUPEFILTER_DEBUG')
        bit = settings.getint('BLOOMFILTER_BIT', 30)
        hash_number = settings.getint('BLOOMFILTER_HASH_NUMBER', 6)
        return cls(server, key=key, debug=debug, bit=bit, hash_number=hash_number)

    @classmethod
    async def from_spider(cls, spider):
        settings = spider.settings
        server = await get_pool('redis')
        dupefilter_key = settings.get("SCHEDULER_DUPEFILTER_KEY",  '%(spider)s:bloomfilter')
        key = dupefilter_key % {'spider': spider.name}
        debug = settings.getbool('DUPEFILTER_DEBUG', False)
        bit = settings.getint('BLOOMFILTER_BIT', 30)
        hash_number = settings.getint('BLOOMFILTER_HASH_NUMBER', 6)
        return cls(server, key=key, debug=debug, bit=bit, hash_number=hash_number)

    async def request_seen(self, request):
        fp = self.request_fingerprint(request)
        if await self.bf.exists(fp):
            return True
        await self.bf.insert(fp)
        return False

    def log(self, request, spider):
        super().log(request, spider)
        spider.crawler.stats.inc_value('bloomfilter/filtered', spider=spider)



# =============调度=======================================
SCHEDULER = 'aioscrapy.core.scheduler.Scheduler'
SCHEDULER_QUEUE_KEY = '%(spider)s:requests'
SCHEDULER_QUEUE_CLASS = 'aioscrapy.core.scheduler.queue.PriorityQueue'
SCHEDULER_PERSIST = True

# 去重
SCHEDULER_DUPEFILTER_KEY = '%(spider)s:dupefilter'
DUPEFILTER_CLASS = 'aioscrapy.core.scheduler.dupefilter.RFPDupeFilter'
# =============调度=======================================


# ===========下载器===================
DOWNLOADER = 'aioscrapy.core.downloader.Downloader'
DOWNLOAD_HANDLERS_BASE = {
    'http': 'aioscrapy.core.downloader.handlers.http.AioHttpDownloadHandler',
    'https': 'aioscrapy.core.downloader.handlers.http.AioHttpDownloadHandler',
}
# ===========下载器===================


# ===========下载中间件 ======================
from scrapy.settings.default_settings import DOWNLOADER_MIDDLEWARES_BASE

DOWNLOADER_MIDDLEWARES_BASE.update({
    # 不需要对response做压缩相关的处理, aiohttp已经处理了
    'scrapy.downloadermiddlewares.httpcompression.HttpCompressionMiddleware': None,
    'scrapy.downloadermiddlewares.retry.RetryMiddleware': None,
    'aioscrapy.downloadermiddlewares.retry.RetryMiddleware': 550,
    'aioscrapy.downloadermiddlewares.tls_ciphers.TLSCiphersMiddleware': 950,
})
# ===========下载中间件 ======================


# ===========item中间件 ======================
ITEM_PROCESSOR = 'aioscrapy.middleware.ItemPipelineManager'
# ===========item中间件 ======================


# =========扩展中间件===============================
EXTENSIONS_BASE = {
    'aioscrapy.extensions.corestats.CoreStats': 0,
    # 'aioscrapy.extensions.telnet.TelnetConsole': 0,
    'aioscrapy.extensions.memusage.MemoryUsage': 0,
    'aioscrapy.extensions.memdebug.MemoryDebugger': 0,
    'aioscrapy.extensions.closespider.CloseSpider': 0,
    # 'aioscrapy.extensions.feedexport.FeedExporter': 0,
    'aioscrapy.extensions.logstats.LogStats': 0,
    'aioscrapy.extensions.spiderstate.SpiderState': 0,
    'aioscrapy.extensions.throttle.AutoThrottle': 0,
}
# =========扩展中间件===============================

# =========爬虫中间件===============================
SPIDER_MIDDLEWARES_BASE = {
    # Engine side
    'aioscrapy.spidermiddlewares.httperror.HttpErrorMiddleware': 50,
    'aioscrapy.spidermiddlewares.offsite.OffsiteMiddleware': 500,
    'aioscrapy.spidermiddlewares.referer.RefererMiddleware': 700,
    'aioscrapy.spidermiddlewares.urllength.UrlLengthMiddleware': 800,
    'aioscrapy.spidermiddlewares.depth.DepthMiddleware': 900,
    # Spider side
}
# =========爬虫中间件===============================




from scrapy import signals
from scrapy.exceptions import DontCloseSpider
from scrapy.spiders import Spider, CrawlSpider

__all__ = ["AioSpider", "AioCrawlSpider"]


class ExtensionMixin:

    @classmethod
    def start(cls, setting_path=None):
        from aioscrapy.crawler import CrawlerProcess
        from aioscrapy.utils.tools import get_project_settings

        settings = get_project_settings()
        if setting_path is not None:
            settings.setmodule(setting_path)
        cp = CrawlerProcess(settings)
        cp.crawl(cls)
        cp.start()

    def spider_idle(self):
        raise DontCloseSpider


class AioSpider(ExtensionMixin, Spider):
    def _set_crawler(self, crawler):
        super()._set_crawler(crawler)
        crawler.signals.connect(self.spider_idle, signal=signals.spider_idle)


class AioCrawlSpider(ExtensionMixin, CrawlSpider):
    def _set_crawler(self, crawler):
        super()._set_crawler(crawler)
        crawler.signals.connect(self.spider_idle, signal=signals.spider_idle)




from aioredis import BlockingConnectionPool, Redis

from .absmanager import ABSManager


class AioRedisManager(ABSManager):
    _clients = {}

    async def create(self, alias: str, params: dict) -> Redis:
        if alias in self._clients:
            return self._clients[alias]
        
        params = params.copy()
        url = params.pop('url', None)
        if url:
            connection_pool = BlockingConnectionPool.from_url(url, **params)
        else:
            connection_pool = BlockingConnectionPool(**params)
        redis = Redis(connection_pool=connection_pool)
        return self._clients.setdefault(alias, redis)

    def get_pool(self, alias: str):
        """获取数据库链接和数据库游标"""
        redis_pool = self._clients.get(alias)
        assert redis_pool is not None, f"redis没有创建该连接池： {alias}"
        return redis_pool

    async def close(self, alias: str):
        """关闭指定redis pool"""
        redis = self._clients.pop(alias, None)
        if redis:
            await redis.close()
            await redis.connection_pool.disconnect()

    async def close_all(self):
        for alias in list(self._clients.keys()):
            await self.close(alias)

    async def from_dict(self, db_args: dict):
        for alias, redis_args in db_args.items():
            await self.create(alias, redis_args)

    async def from_settings(self, settings: "scrapy.settings.Setting"):
        for alias, redis_args in settings.getdict('REDIS_ARGS').items():
            await self.create(alias, redis_args)


redis_manager = AioRedisManager()


if __name__ == '__main__':
    import asyncio

    async def test():
        redis = await redis_manager.create('default', {
                'url': 'redis://:erpteam_redis@192.168.5.216:6381/9',
                'max_connections': 4
            })

        await redis.zadd('key1', {'value': 1})

        async with redis.pipeline(transaction=True) as pipe:
            results, count = await (
                pipe.zrange('key1', 0, 0)
                    .zremrangebyrank('key1', 0, 0)
                    .execute()
            )

        print(results)
        await redis_manager.close_all()


    # asyncio.run(test(), debug=True)
    asyncio.get_event_loop().run_until_complete(test())




from abc import ABCMeta, abstractmethod


class ABSManager(object, metaclass=ABCMeta):

    @property
    @abstractmethod
    def _clients(self):
        return {}

    @abstractmethod
    async def create(self, alias: str, params: dict):
        """创建连接池"""

    @abstractmethod
    def get_pool(self, alias: str):
        """获取指定连接池"""

    @abstractmethod
    async def close_all(self):
        """关闭所有连接池"""

    @abstractmethod
    async def close(self, alias: str):
        """关闭指定连接池"""

    @abstractmethod
    async def from_dict(self, db_args: dict):
        """更具dict实例化连接池管理类"""

    @abstractmethod
    async def from_settings(self, settings: "scrapy.settings.Setting"):
        """更具setting实例化连接池管理类"""

    async def from_crawler(self, crawler):
        return await self.from_settings(crawler.settings)

    def __call__(self, alias):
        return self.get_pool(alias)

    def __getitem__(self, alias):
        return self.get_pool(alias)

    def __getattr__(self, alias):
        return self.get_pool(alias)



"""
Spider Middleware manager

See documentation in docs/topics/spider-middleware.rst
"""
from itertools import islice
from types import AsyncGeneratorType

from scrapy.exceptions import _InvalidOutput
from scrapy.utils.conf import build_component_list

from aioscrapy.utils.tools import async_generator_wrapper
from aioscrapy.utils.tools import call_helper
from .middleware import MiddlewareManager


def _isiterable(possible_iterator):
    return hasattr(possible_iterator, '__iter__')


def _is_async_generator(possible_iterator):
    return isinstance(possible_iterator, AsyncGeneratorType)


def _fname(f):
    return "{}.{}".format(
        f.__self__.__class__.__name__,
        f.__func__.__name__
    )


class SpiderMiddlewareManager(MiddlewareManager):
    component_name = 'spider middleware'

    @classmethod
    def _get_mwlist_from_settings(cls, settings):
        return build_component_list(settings.getwithbase('SPIDER_MIDDLEWARES'))

    def _add_middleware(self, mw):
        super(SpiderMiddlewareManager, self)._add_middleware(mw)
        if hasattr(mw, 'process_spider_input'):
            self.methods['process_spider_input'].append(mw.process_spider_input)
        if hasattr(mw, 'process_start_requests'):
            self.methods['process_start_requests'].appendleft(mw.process_start_requests)
        process_spider_output = getattr(mw, 'process_spider_output', None)
        self.methods['process_spider_output'].appendleft(process_spider_output)
        process_spider_exception = getattr(mw, 'process_spider_exception', None)
        self.methods['process_spider_exception'].appendleft(process_spider_exception)

    async def scrape_response(self, scrape_func, response, request, spider):

        async def process_spider_input(response):
            for method in self.methods['process_spider_input']:
                try:
                    result = await call_helper(method, response=response, spider=spider)
                    if result is not None:
                        msg = "Middleware {} must return None or raise an exception, got {}"
                        raise _InvalidOutput(msg.format(_fname(method), type(result)))
                except _InvalidOutput:
                    raise
                except Exception as exception:
                    iterable_or_exception = await call_helper(scrape_func, exception, request, spider)
                    if iterable_or_exception is exception:
                        raise iterable_or_exception
                    return iterable_or_exception
            return await call_helper(scrape_func, response, request, spider)

        async def _evaluate_iterable(maybe_async_gen, exception_processor_index):
            try:
                # 将所有非AsyncGeneratorType变成AsyncGeneratorType对象
                async_gen = await async_generator_wrapper(maybe_async_gen)
                async for r in async_gen:
                    yield r
            except Exception as ex:
                exception_result = await process_spider_exception(ex, exception_processor_index)
                if isinstance(exception_result, (Exception, BaseException)):
                    raise exception_result
                async for r in _evaluate_iterable(exception_result, exception_processor_index):
                    yield r

        async def process_spider_exception(exception, start_index=0):
            # don't handle _InvalidOutput exception
            if isinstance(exception, _InvalidOutput):
                raise exception
            method_list = islice(self.methods['process_spider_exception'], start_index, None)
            for method_index, method in enumerate(method_list, start=start_index):
                if method is None:
                    continue
                result = await call_helper(method, response=response, exception=exception, spider=spider)
                if _is_async_generator(result) or _isiterable(result):
                    # stop exception handling by handing control over to the
                    # process_spider_output chain if an iterable has been returned
                    return await process_spider_output(result, method_index + 1)
                elif result is None:
                    continue
                else:
                    msg = "Middleware {} must return None or an iterable, got {}"
                    raise _InvalidOutput(msg.format(_fname(method), type(result)))
            raise exception

        async def process_spider_output(result, start_index=0):
            # items in this iterable do not need to go through the process_spider_output
            # chain, they went through it already from the process_spider_exception method

            method_list = islice(self.methods['process_spider_output'], start_index, None)
            for method_index, method in enumerate(method_list, start=start_index):
                if method is None:
                    continue
                try:
                    # might fail directly if the output value is not a generator
                    result = await call_helper(method, response=response, result=result, spider=spider)
                except Exception as ex:
                    exception_result = await process_spider_exception(ex, method_index + 1)
                    if isinstance(exception_result, (Exception, BaseException)):
                        raise
                    return exception_result
                if _is_async_generator(result) or _isiterable(result):
                    result = _evaluate_iterable(result, method_index + 1)
                else:
                    msg = "Middleware {} must return an iterable, got {}"
                    raise _InvalidOutput(msg.format(_fname(method), type(result)))

            return result

        async def process_callback_output(result):
            result = _evaluate_iterable(result, 0)
            return await process_spider_output(result)

        try:
            iterable = await process_spider_input(response)
        except (Exception, BaseException) as exc:
            result = await process_spider_exception(exc)
        else:
            result = await process_callback_output(iterable)
        return result

    async def process_start_requests(self, start_requests, spider):
        return await self._process_chain('process_start_requests', start_requests, spider)



import pickle
import json

from scrapy.utils.python import to_unicode


__all__ = ['PickleCompat', 'JsonCompat']


def _request_byte2str(obj):
    _encoding = obj.get('_encoding', 'utf-8')
    obj.update({
        'body': obj['body'].decode(_encoding),
        'headers': {
            to_unicode(k, encoding=_encoding): to_unicode(b','.join(v), encoding=_encoding)
            for k, v in obj['headers'].items()
        }
    })
    return obj


class PickleCompat:
    @staticmethod
    def loads(s):
        return pickle.loads(s)

    @staticmethod
    def dumps(obj):
        return pickle.dumps(obj, protocol=-1)


class JsonCompat:
    @staticmethod
    def loads(s):
        return json.loads(s)

    @staticmethod
    def dumps(obj):
        return json.dumps(_request_byte2str(obj))



import logging

from ._aioredis import redis_manager
from ._aiomysql import mysql_manager

logger = logging.getLogger(__name__)

__all__ = ['db_manager', 'get_pool', 'get_manager']


db_manager_map = {
    'mysql': mysql_manager,
    'redis': redis_manager,
}


class DBManager:

    @staticmethod
    def get_manager(db_type):
        manager = db_manager_map.get(db_type)
        assert manager is not None, f"暂时不支持该类型数据库：{db_type}"
        return manager

    def get_pool(self, db_type, alias='default'):
        manager = self.get_manager(db_type)
        return manager.get_pool(alias)

    @staticmethod
    async def close_all():
        for manager in db_manager_map.values():
            await manager.close_all()

    @staticmethod
    async def from_dict(db_args: dict):
        for db_type, args in db_args.items():
            manager = db_manager_map.get(db_type)
            if manager is None:
                logger.warning(f'Not support db type: {db_type}; Only {", ".join(db_manager_map.keys())} supported')
            await manager.from_dict(args)

    @staticmethod
    async def from_settings(settings: "scrapy.settings.Setting"):
        for manager in db_manager_map.values():
            await manager.from_settings(settings)

    async def from_crawler(self, crawler):
        return await self.from_settings(crawler.settings)

    def __getattr__(self, db_type: str):
        if db_type not in db_manager_map:
            raise AttributeError(f'Not support db type: {db_type}')
        return db_manager_map[db_type]


db_manager = DBManager()
get_manager = db_manager.get_manager
get_pool = db_manager.get_pool



# _*_ coding: utf-8 _*_

""" 响应类 """
import re

from scrapy.http import Response
from scrapy.http import TextResponse as STextResponse

__all__ = ['Response', 'TextResponse']


class TextResponse(STextResponse):
    def __init__(self, *args, **kwargs):
        self.cookies = self.deal_cookies(kwargs.pop("cookies", None))
        super().__init__(*args, **kwargs)

    @staticmethod
    def deal_cookies(cookies_raw):
        cookies = {}
        if cookies_raw is None:
            return cookies
        cookies_str = cookies_raw.output()
        for cookie in re.findall(r'Set-Cookie: (.*?)=(.*?); Domain', cookies_str, re.S):
            cookies[cookie[0]] = cookie[1]
        return cookies



import logging
import pprint
import warnings
import asyncio
import signal
import sys

from zope.interface.exceptions import DoesNotImplement

try:
    # zope >= 5.0 only supports MultipleInvalid
    from zope.interface.exceptions import MultipleInvalid
except ImportError:
    MultipleInvalid = None

from zope.interface.verify import verifyClass
from scrapy import signals, Spider
from scrapy.settings import overridden_settings
from scrapy.utils.log import (
    get_scrapy_root_handler,
    install_scrapy_root_handler,
    LogCounterHandler,
    configure_logging,
)
from scrapy.utils.misc import load_object
from scrapy.interfaces import ISpiderLoader
from scrapy.exceptions import ScrapyDeprecationWarning

from aioscrapy.utils.tools import async_generator_wrapper
from aioscrapy.middleware import ExtensionManager
from aioscrapy.core.engine import ExecutionEngine
from aioscrapy.settings import AioSettings
from aioscrapy.signalmanager import SignalManager
from aioscrapy.utils.ossignal import install_shutdown_handlers, signal_names

logger = logging.getLogger(__name__)


class Crawler:

    def __init__(self, spidercls, *args, settings=None, **kwargs):
        if isinstance(spidercls, Spider):
            raise ValueError('The spidercls argument must be a class, not an object')

        if isinstance(settings, dict) or settings is None:
            settings = AioSettings(settings)

        self.spidercls = spidercls
        self.settings = settings.copy()
        self.spidercls.update_settings(self.settings)

        self.signals = SignalManager(self)
        self.stats = load_object(self.settings['STATS_CLASS'])(self)

        handler = LogCounterHandler(self, level=self.settings.get('LOG_LEVEL'))
        logging.root.addHandler(handler)

        d = dict(overridden_settings(self.settings))
        logger.info("Overridden settings:\n%(settings)s",
                    {'settings': pprint.pformat(d)})

        if get_scrapy_root_handler() is not None:
            install_scrapy_root_handler(self.settings)
        self.__remove_handler = lambda: logging.root.removeHandler(handler)
        self.signals.connect(self.__remove_handler, signals.engine_stopped)

        lf_cls = load_object(self.settings['LOG_FORMATTER'])
        self.logformatter = lf_cls.from_crawler(self)
        self.extensions = ExtensionManager.from_crawler(self)
        self.settings.freeze()
        self.crawling = False
        self.spider = self._create_spider(*args, **kwargs)
        self.engine = None

    async def crawl(self):
        if self.crawling:
            raise RuntimeError("Crawling already taking place")
        self.crawling = True

        try:
            self.engine = self._create_engine()
            start_requests = await async_generator_wrapper(self.spider.start_requests())
            await self.engine.start(self.spider, start_requests)
        except Exception as e:
            logger.exception(e)
            self.crawling = False
            if self.engine is not None:
                await self.engine.close()
            raise e

    def _create_spider(self, *args, **kwargs):
        return self.spidercls.from_crawler(self, *args, **kwargs)

    def _create_engine(self):
        return ExecutionEngine(self, self.stop)

    async def stop(self):
        """Starts a graceful stop of the crawler and returns a deferred that is
        fired when the crawler is stopped."""
        if self.crawling:
            self.crawling = False
            await self.engine.stop()


class CrawlerRunner:

    crawlers = property(
        lambda self: self._crawlers,
        doc="Set of :class:`crawlers <scrapy.crawler.Crawler>` started by "
            ":meth:`crawl` and managed by this class."
    )

    @staticmethod
    def _get_spider_loader(settings):
        """ Get SpiderLoader instance from settings """
        cls_path = settings.get('SPIDER_LOADER_CLASS')
        loader_cls = load_object(cls_path)
        excs = (DoesNotImplement, MultipleInvalid) if MultipleInvalid else DoesNotImplement
        try:
            verifyClass(ISpiderLoader, loader_cls)
        except excs:
            warnings.warn(
                'SPIDER_LOADER_CLASS (previously named SPIDER_MANAGER_CLASS) does '
                'not fully implement scrapy.interfaces.ISpiderLoader interface. '
                'Please add all missing methods to avoid unexpected runtime errors.',
                category=ScrapyDeprecationWarning, stacklevel=2
            )
        return loader_cls.from_settings(settings.frozencopy())

    def __init__(self, settings=None):
        if isinstance(settings, dict) or settings is None:
            settings = AioSettings(settings)
        self.settings = settings
        self.spider_loader = self._get_spider_loader(settings)
        self._crawlers = set()
        self._active = set()
        self.bootstrap_failed = False

    @property
    def spiders(self):
        warnings.warn("CrawlerRunner.spiders attribute is renamed to "
                      "CrawlerRunner.spider_loader.",
                      category=ScrapyDeprecationWarning, stacklevel=2)
        return self.spider_loader

    def crawl(self, crawler_or_spidercls, *args, **kwargs):
        if isinstance(crawler_or_spidercls, Spider):
            raise ValueError(
                'The crawler_or_spidercls argument cannot be a spider object, '
                'it must be a spider class (or a Crawler object)')
        crawler = self.create_crawler(crawler_or_spidercls, *args, **kwargs)
        self.crawlers.add(crawler)
        return crawler

    def crawl_soon(self, crawler_or_spidercls, *args, **kwargs):
        crawler = self.crawl(crawler_or_spidercls, *args, **kwargs)
        self.active_crawler(crawler)

    def active_crawler(self, crawler):
        task = asyncio.create_task(crawler.crawl())
        self._active.add(task)

        def _done(result):
            self.crawlers.discard(crawler)
            self._active.discard(task)
            self.bootstrap_failed |= not getattr(crawler, 'spider', None)
            return result

        task.add_done_callback(_done)

    def create_crawler(self, crawler_or_spidercls, *args, **kwargs):
        if isinstance(crawler_or_spidercls, Spider):
            raise ValueError(
                'The crawler_or_spidercls argument cannot be a spider object, '
                'it must be a spider class (or a Crawler object)')
        if isinstance(crawler_or_spidercls, Crawler):
            return crawler_or_spidercls
        return self._create_crawler(crawler_or_spidercls, *args, **kwargs)

    def _create_crawler(self, spidercls, *args, **kwargs):
        if isinstance(spidercls, str):
            spidercls = self.spider_loader.load(spidercls)
        settings = kwargs.pop('settings', self.settings)
        return Crawler(spidercls, *args, settings=settings, **kwargs)

    async def stop(self):
        return await asyncio.gather(*[c.stop() for c in list(self.crawlers)])


class CrawlerProcess(CrawlerRunner):

    def __init__(self, settings=None, install_root_handler=True):
        super().__init__(settings)
        install_shutdown_handlers(self._signal_shutdown)
        configure_logging(self.settings, install_root_handler)

    def _signal_shutdown(self, signum, _):
        install_shutdown_handlers(self._signal_kill)
        signame = signal_names[signum]
        logger.info("Received %(signame)s, shutting down gracefully. Send again to force ",
                    {'signame': signame})
        asyncio.create_task(self._graceful_stop_reactor())

    def _signal_kill(self, signum, _):
        install_shutdown_handlers(signal.SIG_IGN)
        signame = signal_names[signum]
        logger.info('Received %(signame)s twice, forcing unclean shutdown',
                    {'signame': signame})
        asyncio.create_task(self._stop_reactor())

    async def run(self):
        for crawler in self.crawlers:
            self.active_crawler(crawler)
        while self._active:
            await asyncio.gather(*self._active)
        await self.recycle_db_connect()

    def start(self):
        if sys.platform.startswith('win'):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        else:
            try:
                import uvloop
                asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            except ImportError:
                pass
        asyncio.run(self.run())

    async def _graceful_stop_reactor(self):
        await self.stop()
        await self.recycle_db_connect()

    async def _stop_reactor(self):
        try:
            await self.recycle_db_connect()
        finally:
            asyncio.get_event_loop().stop()

    async def recycle_db_connect(self):
        # 回收所以的链接
        if not len(self._active):
            from aioscrapy.db import db_manager
            await db_manager.close_all()



import asyncio
import logging
import ssl
import sys

from scrapy.http import Headers
from aioscrapy.https import TextResponse
import aiohttp

logger = logging.getLogger(__name__)


class AioHttpDownloadHandler:
    session = None

    def __init__(self, settings):
        self.settings = settings
        self.aiohttp_client_session_args = settings.get('AIOHTTP_CLIENT_SESSION_ARGS', {})
        self.verify_ssl = self.settings.get("VERIFY_SSL")

    @classmethod
    def from_settings(cls, settings):
        return cls(settings)

    def get_session(self, *args, **kwargs):
        if self.session is None:
            self.session = aiohttp.ClientSession(*args, **kwargs)
        return self.session

    async def download_request(self, request, spider):
        kwargs = {
            'verify_ssl': request.meta.get('verify_ssl', self.verify_ssl),
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
            'data': request.body or None
        }

        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        if isinstance(headers, Headers):
            headers = headers.to_unicode_dict()
        kwargs['headers'] = headers

        ssl_ciphers = request.meta.get('TLS_CIPHERS')
        if ssl_ciphers:
            context = ssl.create_default_context()
            context.set_ciphers(ssl_ciphers)
            kwargs['ssl'] = context

        proxy = request.meta.get("proxy")
        if proxy:
            kwargs["proxy"] = proxy
            logger.info(f"使用代理{proxy}抓取: {request.url}")

            async with aiohttp.ClientSession(**self.aiohttp_client_session_args) as session:
                async with session.request(request.method, request.url, **kwargs) as response:
                    content = await response.read()

        # 当不用代理的时候，session不关闭
        else:
            session = self.get_session(**self.aiohttp_client_session_args)
            async with session.request(request.method, request.url, **kwargs) as response:
                content = await response.read()

        return TextResponse(str(response.url),
                            status=response.status,
                            headers=response.headers,
                            body=content,
                            cookies=response.cookies)

    async def close(self):
        if self.session is not None:
            await self.session.close()

        # Wait 250 ms for the underlying SSL connections to close
        # https://docs.aiohttp.org/en/latest/client_advanced.html#graceful-shutdown
        await asyncio.sleep(0.250)
